# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
from typing import Any
import yaml


DEFAULT_CONFIG = "kithara/config/default.yaml"


def load_config(argv) -> dict[str, Any]:
  """Loads the YAML config from a file with a given name."""
  config_name = DEFAULT_CONFIG
  if len(argv) > 1:
    config_name = argv[1]

  with open(config_name, "r", encoding="utf-8") as yaml_file:
    config = yaml.safe_load(yaml_file)

  # Some float parameters like learning_rate could be parsed
  # as strings instead of floats, so we need to explicitly
  # convert them.
  if 'learning_rate' in config:
    config['learning_rate'] = float(config['learning_rate'])


  print(f"Parsed config: {config}")
  return config
